"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import os
import shutil
from logging import Logger
from json.decoder import JSONDecodeError

from Setter.Common.AuxiliaryFunctions import AuxiliaryFunctions
from Setter.Common.Boards import Boards
from Setter.Common.Command import Command
from Setter.Common.PasswordGenerator import PasswordGenerator
from Setter.Enums.Interfaces import InterfaceCodes, InterfaceNames
from Setter.Exceptions.BackupException import BackupException
from Setter.Exceptions.IOErrorException import IOErrorException
from Setter.Exceptions.JSONErrorException import JSONErrorException
from Setter.Logging import Logging
from Setter.MosquittoConfig import MosquittoConfig


class DaemonConfig:
	"""
	IQRF Gateway Daemon configuration class.

	Methods
	-------
	backup() -> None:
		Backup daemon configuration.
	restore() -> None:
		Restore daemon configuration from backup.
	configure_interface(product: str, interface: InterfaceCodes) -> None:
		Configure IQRF interface specified product.
	configure_splitter(instance_id: str) -> None:
		Configure JSON splitter component instance.
	configure_ide(instance_id: str) -> None:
		Configure IDE counterpart component instance.
	configure_mqtt(instance_id: str) -> None:
		Configure MQTT messaging component instance.
	"""

	__backup_dir: str = '/tmp/iqrf-gateway-daemon/'
	__config_dir: str = '/etc/iqrf-gateway-daemon/'
	__name: str = 'IQRF Gateway Daemon'
	__logger: Logger = Logging.get_logger()

	@classmethod
	def backup(cls):
		"""
		Backup daemon configuration.

		Current configuration is stored to tmp.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		BackupException
			raised if configuration backup failed
		"""
		cls.__logger.info('Backing up %s to tmp', cls.__name)
		os.makedirs(cls.__backup_dir, exist_ok=True)
		try:
			shutil.copy2(cls.__config_dir + 'config.json', cls.__backup_dir)
			shutil.copy2(cls.__config_dir + 'iqrf__IqrfCdc.json', cls.__backup_dir)
			shutil.copy2(cls.__config_dir + 'iqrf__IqrfSpi.json', cls.__backup_dir)
			shutil.copy2(cls.__config_dir + 'iqrf__IqrfUart.json', cls.__backup_dir)
			shutil.copy2(cls.__config_dir + 'iqrf__IdeCounterpart.json', cls.__backup_dir)
			shutil.copy2(cls.__config_dir + 'iqrf__JsonSplitter.json', cls.__backup_dir)
			shutil.copy2(cls.__config_dir + 'iqrf__MqttMessaging.json', cls.__backup_dir)
		except IOError as io_error:
			error_str = f'Failed to backup {cls.__name} configuration: {io_error.strerror}.'
			cls.__logger.error(error_str)
			AuxiliaryFunctions.rm_rf(cls.__backup_dir)
			raise BackupException(error_str) from io_error

	@classmethod
	def restore(cls):
		"""
		Restore daemon configuration from backup.

		Backed up configuration from tmp is used to restore daemon configuration.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		cls.__logger.info('Restoring %s from tmp', cls.__name)
		files = 'config.json,iqrf__IqrfCdc.json,iqrf__IqrfSpi.json,iqrf__IqrfUart.json,' \
			'iqrf__IdeCounterpart.json,iqrf__JsonSplitter.json,iqrf__MqttMessaging.json'
		command = f'cp -pf {cls.__backup_dir}{{{files}}} {cls.__config_dir}'
		result = Command.execute(command)
		if result.returncode != 0:
			error_str = f'Failed to restore {cls.__name} configuration: {result.stderr}.'
			cls.__logger.error(error_str)
		try:
			shutil.copy2(cls.__backup_dir + 'config.json', cls.__config_dir)
			shutil.copy2(cls.__backup_dir + 'iqrf__IqrfCdc.json', cls.__config_dir)
			shutil.copy2(cls.__backup_dir + 'iqrf__IqrfSpi.json', cls.__config_dir)
			shutil.copy2(cls.__backup_dir + 'iqrf__IqrfUart.json', cls.__config_dir)
			shutil.copy2(cls.__backup_dir + 'iqrf__IdeCounterpart.json', cls.__config_dir)
			shutil.copy2(cls.__backup_dir + 'iqrf__JsonSplitter.json', cls.__config_dir)
			shutil.copy2(cls.__backup_dir + 'iqrf__MqttMessaging.json', cls.__config_dir)
		except IOError as io_error:
			error_str = f'Failed to restore {cls.__name} configuration: {io_error.strerror}.'
			cls.__logger.error(error_str)
		AuxiliaryFunctions.rm_rf(cls.__backup_dir)

	@classmethod
	def configure_interface(cls, product: str, interface: InterfaceCodes) -> None:
		"""
		Configure IQRF interface specified product.

		Specified interface is configured and enabled, other interfaces are disabled.

		Parameters
		----------
		product: str
			product identifier
		interface: InterfaceCodes
			interface to configure

		Returns
		-------
		None

		Exceptions
		----------
		IOErrorException
			raised if configuration files cannot be read or written to
		JSONDecodeError
			raised if configuration file(s) have invalid json syntax
		"""
		cls.__logger.info('Configuring %s for %s and %s interface', cls.__name, product, interface)
		board_config = Boards.get_model_config(product, interface)
		cdc, spi, uart = False, False, False

		try:
			file_name = cls.__config_dir + 'config.json'
			with open(file_name, mode='r', encoding='utf-8') as file:
				daemon_config = json.load(file)

			if interface == InterfaceCodes.CDC:
				for i in range(len(daemon_config['components'])):
					if cdc and spi and uart:
						break
					if daemon_config['components'][i]['name'] == InterfaceNames.CDC:
						daemon_config['components'][i]['enabled'] = True
						cdc = True
					elif daemon_config['components'][i]['name'] == InterfaceNames.SPI:
						daemon_config['components'][i]['enabled'] = False
						spi = True
					elif daemon_config['components'][i]['name'] == InterfaceNames.UART:
						daemon_config['components'][i]['enabled'] = False
						uart = True
				return
			if interface == InterfaceCodes.SPI:
				for i in range(len(daemon_config['components'])):
					if cdc and spi and uart:
						break
					if daemon_config['components'][i]['name'] == InterfaceNames.CDC:
						daemon_config['components'][i]['enabled'] = False
						cdc = True
					elif daemon_config['components'][i]['name'] == InterfaceNames.SPI:
						daemon_config['components'][i]['enabled'] = True
						spi = True
					elif daemon_config['components'][i]['name'] == InterfaceNames.UART:
						daemon_config['components'][i]['enabled'] = False
						uart = True
			if interface == InterfaceCodes.UART:
				for i in range(len(daemon_config['components'])):
					if cdc and spi and uart:
						break
					if daemon_config['components'][i]['name'] == InterfaceNames.CDC:
						daemon_config['components'][i]['enabled'] = False
						cdc = True
					elif daemon_config['components'][i]['name'] == InterfaceNames.SPI:
						spi = True
						daemon_config['components'][i]['enabled'] = False
					elif daemon_config['components'][i]['name'] == InterfaceNames.UART:
						daemon_config['components'][i]['enabled'] = True
						uart = True

			with open(file_name, mode='w', encoding='utf-8') as file:
				json.dump(daemon_config, file, indent=4)

			file_name = cls.__config_dir + ('iqrf__IqrfSpi.json' if interface == InterfaceCodes.SPI else 'iqrf__IqrfUart.json')
			with open(file_name, mode='w', encoding='utf-8') as file:
				json.dump(board_config.to_daemon(), file, indent=4)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error
		except JSONDecodeError as json_error:
			error_str = Logging.json_error(json_error, file_name)
			cls.__logger.error(error_str)
			raise JSONErrorException(error_str) from json_error
		AuxiliaryFunctions.rm_rf(cls.__backup_dir)

	@classmethod
	def configure_splitter(cls, instance_id: str) -> None:
		"""
		Configure JSON splitter component instance.

		Parameters
		----------
		instance_id: str
			instance identifier

		Returns
		-------
		None

		Exceptions
		----------
		IOErrorException
			raised if configuration cannot be read or written to
		JSONDecodeError
			raised if configuration file has invalid json syntax
		"""
		cls.__logger.info('Configuring %s Splitter', cls.__name)
		file_name = cls.__config_dir + 'iqrf__JsonSplitter.json'
		try:
			with open(file_name, mode='r', encoding='utf-8') as file:
				splitter_config = json.load(file)

			splitter_config['insId'] = instance_id

			with open(file_name, mode='w', encoding='utf-8') as file:
				json.dump(splitter_config, file, indent=4)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error
		except JSONDecodeError as json_error:
			error_str = Logging.json_error(json_error, file_name)
			cls.__logger.error(error_str)
			raise JSONErrorException(error_str) from json_error

	@classmethod
	def configure_ide(cls, instance_id: str) -> None:
		"""
		Configure IDE counterpart component instance.

		Configure IDE counterpart identification name and byte.

		Parameters
		----------
		instance_id: str
			instance identifier

		Returns
		-------
		None

		Exceptions
		----------
		IOErrorException
			raised if configuration cannot be read or written to
		JSONDecodeError
			raised if configuration file has invalid json syntax
		"""
		cls.__logger.info('Configuring %s IDE Counterpart', cls.__name)
		file_name = cls.__config_dir + 'iqrf__IdeCounterpart.json'
		try:
			with open(file_name, mode='r', encoding='utf-8') as file:
				ide_config = json.load(file)

			ide_config['gwIdentName'] = instance_id
			ide_config['gwIdentModeByte'] = 34

			with open(file_name, mode='w', encoding='utf-8') as file:
				json.dump(ide_config, file, indent=4)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error
		except JSONDecodeError as json_error:
			error_str = Logging.json_error(json_error, file_name)
			cls.__logger.error(error_str)
			raise JSONErrorException(error_str) from json_error

	@classmethod
	def configure_mqtt(cls, instance_id: str) -> None:
		"""
		Configure MQTT messaging component instance.

		Parameters
		----------
		instance_id: str
			instance identifier

		Returns
		-------
		None

		Exceptions
		----------
		IOErrorException
			raised if configuration cannot be read or written to
		JSONDecodeError
			raised if configuration file has invalid json syntax
		"""
		cls.__logger.info('Configuring %s MQTT messaging', cls.__name)
		try:
			file_name = cls.__config_dir + 'iqrf__MqttMessaging.json'
			with open(file_name, mode='r', encoding='utf-8') as file:
				mqtt_config = json.load(file)

			username = 'iqrf-gateway-daemon'
			password = PasswordGenerator.generate()
			MosquittoConfig.create_user(username, password)

			mqtt_config['ClientId'] = instance_id
			mqtt_config['User'] = username
			mqtt_config['Password'] = password
			mqtt_config['TopicRequest'] = f'gateway/{instance_id}/iqrf/requests'
			mqtt_config['TopicResponse'] = f'gateway/{instance_id}/iqrf/responses'

			with open(file_name, mode='w', encoding='utf-8') as file:
				json.dump(mqtt_config, file, indent=4)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error
		except JSONDecodeError as json_error:
			error_str = Logging.json_error(json_error, file_name)
			cls.__logger.error(error_str)
			raise JSONErrorException(error_str) from json_error
