"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from logging import Logger

from Setter.Common.Command import Command
from Setter.Exceptions.CommandException import CommandException
from Setter.Exceptions.SystemErrorException import SystemErrorException
from Setter.Logging import Logging


class GrafanaConfig:
	"""
	Grafana configuration class.

	Methods
	-------
	configure() -> None:
		Configure Grafana, set correct permissions and dashboards.
	configure_access(password: str) -> None:
		Configure access password.
	"""

	__logger: Logger = Logging.get_logger()

	@classmethod
	def configure(cls) -> None:
		"""
		Configure Grafana, set correct permissions and dashboards.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if permission or owner change command fails
		SystemErrorException
			raised if file manipulation fails
		"""
		cls.__logger.info('Configuring Grafana.')
		result = Command.execute('chown -R root:grafana /etc/grafana')
		if result.returncode != 0:
			error_str = f'Failed to change owner of grafana files: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

		result = Command.execute('chown -R grafana:grafana /var/lib/grafana')
		if result.returncode != 0:
			error_str = f'Failed to change owner of grafana lib files: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

		sample_dashboard_path = '/etc/grafana/provisioning/dashboards/sample.yaml'
		if os.path.exists(sample_dashboard_path):
			try:
				os.remove(sample_dashboard_path)
			except OSError as os_error:
				error_str = f'Failed to remove sample dashboard: {os_error.strerror}.'
				cls.__logger.error(error_str)
				raise SystemErrorException(error_str) from os_error

		try:
			os.chmod('/etc/grafana/grafana.ini', 0o640)
			os.chmod('/etc/grafana/provisioning/dashboards/iqaros.yaml', 0o660)
			os.chmod('/var/lib/grafana/grafana.db', 0o640)
		except OSError as os_error:
			error_str = f'Failed to change permissions of grafana file {os_error.filename}: {os_error.strerror}.'
			cls.__logger.error(error_str)
			raise SystemErrorException(error_str) from os_error

	@classmethod
	def configure_access(cls, password: str) -> None:
		"""
		Configure access password.

		Parameters
		----------
		password: str
			Password

		Returns
		-------
		None
		"""
		command = f'grafana-cli admin reset-admin-password {password}'
		result = Command.execute(command, True)
		if result.returncode != 0:
			error_str = f'Failed to change grafana password: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)
